import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.TimeUnit;

public class Socket {
	
	public static final int MAX_SIZE_OF_INCOMING_MESSAGE = 1024;
	
	public static final int UNABLE_TO_OBTAIN_IP_ADDRESS_OF_LOCAL_HOST = -1;
	public static final int UNABLE_TO_SEND_DATAGRAM_PACKET = -2;
	public static final int UNABLE_TO_RECEIVE_DATAGRAM_PACKET = -3;
	public static final int UNABLE_TO_OPEN_DATAGRAM_SOCKET = -4;
	public static final int UNABLE_TO_SET_SOCKET_TIMEOUT = -5;
	private static final int UNABLE_TO_SLEEP = -6;

	public static final int SOCKET_TIMEOUT = 50;
	public static final int CLOSE_SOCKET_DELAY = SOCKET_TIMEOUT * 3;
	
	private DatagramSocket mySocket;
	private int myPortNumber;
	private InetAddress myIPAddress;

	private Thread receiveThread;

	private ConcurrentLinkedQueue<DatagramPacket> packetQueue;
	
	private boolean receiveThreadMustExit = false;
	
	public Socket(int myPortNumber) {
		this.myPortNumber = myPortNumber;
		
		try {
			this.myIPAddress = InetAddress.getLocalHost();
			System.out.println("My IP Address = " + myIPAddress.getHostAddress());
		} catch (UnknownHostException e) {
			e.printStackTrace();
			System.exit(UNABLE_TO_OBTAIN_IP_ADDRESS_OF_LOCAL_HOST);
		}
		
		try {
			this.mySocket = new DatagramSocket(this.myPortNumber, this.myIPAddress);
		} catch (SocketException e) {
			e.printStackTrace();
			System.exit(UNABLE_TO_OPEN_DATAGRAM_SOCKET);
		}
		
		this.packetQueue = new ConcurrentLinkedQueue<DatagramPacket>();
		
		// create the receive thread
		this.receiveThread = new Thread(
				new Runnable() {
					@Override
					public void run() {
						receiveThread();
					}
				});
		this.receiveThread.setName("Receive Thread for Port = " + this.myPortNumber);
		this.receiveThread.start();
	}

	
	
	
	public void send(String message, InetAddress destinationIPAddress, int destinationPort) {
		
		byte[] buffer = message.getBytes();
		
		DatagramPacket sendPacket = new DatagramPacket(buffer, 
				buffer.length, 
				destinationIPAddress, 
				destinationPort);
		
		try {
			mySocket.send(sendPacket);
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(UNABLE_TO_SEND_DATAGRAM_PACKET);
		}
	}
	
	
	
	public DatagramPacket receive() {
		return this.packetQueue.poll();
	}
	
	public void receiveThread() {
		System.out.println("Receive Thread is Starting!!!");
		
		try {
			this.mySocket.setSoTimeout(SOCKET_TIMEOUT);
		} catch (SocketException e1) {
			e1.printStackTrace();
			System.exit(UNABLE_TO_SET_SOCKET_TIMEOUT);
		}
		
		do {
			byte[] inBuffer = new byte[MAX_SIZE_OF_INCOMING_MESSAGE];
			
			for ( int i = 0 ; i < inBuffer.length ; i++ ) {
				inBuffer[i] = ' ';
			}
			
			DatagramPacket inPacket = new DatagramPacket(inBuffer, inBuffer.length);
			
			try {
				mySocket.receive(inPacket);
				this.packetQueue.add(inPacket);
			} catch (SocketTimeoutException ste) {
				// nothing to do.
			} catch (IOException e) {
				e.printStackTrace();
				System.exit(UNABLE_TO_RECEIVE_DATAGRAM_PACKET);
			}
		} while(!this.receiveThreadMustExit);
		
		System.out.println("Receive Thread is Stopping!!!");
	}
	
	
	public void close() {
		this.receiveThreadMustExit = true;
		
		try {
			TimeUnit.MILLISECONDS.sleep(CLOSE_SOCKET_DELAY);
		} catch (InterruptedException e) {
			e.printStackTrace();
			System.exit(UNABLE_TO_SLEEP);
		}
		
		this.mySocket.close();
		System.out.println("Socket has been closed");
	}
	
	
	
	
	public static void main(String[] args) {
		Socket s = new Socket(64000);
	}
	
	
}
